import jax.numpy as jnp
from jax.nn import softplus


class Transform:
    def __init__(self, name="Transformation"):
        self.name = name

    @staticmethod
    def forward(x: jnp.ndarray) -> jnp.ndarray:
        raise NotImplementedError

    @staticmethod
    def backward(x: jnp.ndarray) -> jnp.ndarray:
        raise NotImplementedError


class Softplus(Transform):
    def __init__(self):
        super().__init__(name='Softplus')

    @staticmethod
    def forward(x: jnp.ndarray) -> jnp.ndarray:
        return jnp.log(jnp.exp(x) - 1.)

    @staticmethod
    def backward(x: jnp.ndarray) -> jnp.ndarray:
        return softplus(x)


class Identity(Transform):
    def __init__(self):
        super().__init__(name='Identity')

    @staticmethod
    def forward(x: jnp.ndarray) -> jnp.ndarray:
        return x

    @staticmethod
    def backward(x: jnp.ndarray) -> jnp.ndarray:
        return x



import jax.numpy as jnp
from objax import Module
from typing import Callable, Optional
from jax import vmap
from .parameters import Parameter


class Kernel(Module):
    """
    Base class for all kernel functions. By inheriting the `Module` class from Objax, seamless interaction with model parameters is provided.
    """
    def __init__(self, name: Optional[str] = None):
        """
        Args:
            name: Optional naming of the kernel.
        """
        self.name = name
        self.spectral = False

    @staticmethod
    def gram(func: Callable, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
        """
        Compute the kernel's gram matrix given two, possibly identical, Jax arrays.

        Args:
            func: The kernel function to be called for any two values in x and y.
            x: An NxD vector of inputs.
            y: An MXE vector of inputs.

        Returns:
            An NxM gram matrix.
        """
        mapx1 = vmap(lambda x, y: func(x=x, y=y),
                     in_axes=(0, None),
                     out_axes=0)
        mapx2 = vmap(lambda x, y: mapx1(x, y), in_axes=(None, 0), out_axes=1)
        return mapx2(x, y)

    @staticmethod
    def dist(x: jnp.array, y: jnp.array) -> float:
        """
        Compute the squared distance matrix between two inputs.

        Args:
            x: A 1xN vector
            y: A 1xM vector

        Returns:
            A float value quantifying the distance between x and y.
        """
        return jnp.sum((x - y)**2)

    def __call__(self, x: jnp.ndarray, y: jnp.ndarray):
        raise NotImplementedError


class Stationary(Kernel):
    """
    A base class for stationary kernels. That is, kernels whereby the Gram matrix's values are invariant to the value of the inputs, and instead depend only on the distance between the inputs.
    """
    def __init__(self,
                 lengthscale: Optional[jnp.ndarray] = jnp.array([1.]),
                 variance: Optional[jnp.ndarray] = jnp.array([1.]),
                 name: Optional[str] = "Stationary"):
        """
        Args:
            lengthscale: The initial value of the kernel's lengthscale value. The value of this parameter controls the horizontal magnitude of the kernel's resultant values.
            variance: The inital value of the kernel's variance. This value controls the kernel's vertical amplitude.
            name: Optional argument to name the kernel.
        """
        super().__init__(name=name)
        self.lengthscale = Parameter(lengthscale)
        self.variance = Parameter(variance)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class RBF(Stationary):
    """
    The radial basis function kernel.
    """
    def __init__(self,
                 lengthscale: jnp.ndarray = jnp.array([1.]),
                 variance: jnp.ndarray = jnp.array([1.]),
                 name: str = "RBF"):
        """
        Args:
            lengthscale: The initial value of the kernel's lengthscale value. The value of this parameter controls the horizontal magnitude of the kernel's resultant values.
            variance: The initial value of the kernel's variance. This value controls the kernel's vertical amplitude.
            name: Optional argument to name the kernel.
        """
        super().__init__(lengthscale=lengthscale, variance=variance, name=name)

    def feature_map(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
        """
        Compute the RBF specific function
        """
        # :math:`k(x,y)=\sigma^2 \exp\left( \frac{-0.5 \tau}{2\ell^2}\right) ` where  :math:`\tau = \lVert x-y \rVert_{2}^{2}`.
        ell = self.lengthscale.untransform
        sigma = self.variance.untransform
        tau = self.dist(x / ell, y / ell)
        return sigma * jnp.exp(-0.5 * tau)

    def __call__(self, X: jnp.ndarray, Y: jnp.ndarray) -> jnp.ndarray:
        return self.gram(self.feature_map, X, Y).squeeze()



from jax.config import config
config.update("jax_enable_x64", True)
from .gps.priors import Prior
from .kernel import RBF
from .likelihoods import Gaussian
from .mean_functions import ZeroMean
from .utilities import save, load


__version__ = "0.1.0"



from objax import Module
import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular
from .parameters import Parameter
from typing import Optional
from tensorflow_probability.substrates.jax import distributions as tfd


class Likelihood(Module):
    """
    Base class for all likelihood functions. By inheriting the `Module` class from Objax, seamless interaction with
    model parameters is provided.
    """
    def __init__(self, name: Optional[str] = "Likelihood"):
        """
        Args:
            name: Optional naming of the likelihood.
        """
        self.name = name

    def __call__(self):
        raise NotImplementedError


class Gaussian(Likelihood):
    """
    Gaussian likelihood function 
    """
    def __init__(self, noise: jnp.array = jnp.array([1.0]), name="Gaussian"):
        super().__init__(name=name)
        self.noise = Parameter(noise)

    def log_likelihood(self, x: jnp.ndarray, mu: jnp.ndarray, L: jnp.ndarray):
        delta = x - mu
        alpha = solve_triangular(L, delta, lower=True)
        L_diag = jnp.sum(jnp.log(jnp.diag(L)))
        ll = -0.5 * jnp.sum(jnp.square(alpha), axis=0)
        ll -= 0.5 * jnp.log(2 * jnp.pi)
        ll -= L_diag
        return ll

    def __call__(self):
        raise NotImplementedError


class Bernoulli(Likelihood):
    def __init__(self):
        super().__init__(name="Bernoulli")
        self.random_variable = tfd.ProbitBernoulli

    def log_density(self, F, Y):
        return self.random_variable(F).log_prob(Y)

    def mean(self, probs):
        return self.random_variable()



from setuptools import setup


def parse_requirements_file(filename):
    with open(filename, encoding="utf-8") as fid:
        requires = [l.strip() for l in fid.readlines() if l]
    return requires


setup(
    name='GPJax',
    version='0.1.0',
    author='Thomas Pinder',
    author_email='t.pinder2@lancaster.ac.uk',
    packages=['gpjax'],  #, 'package_name.test'],
    # scripts=['bin/script1', 'bin/script2'],
    # url='http://pypi.python.org/pypi/PackageName/',
    license='LICENSE',
    description=
    'Building blocks of Gaussian processes. Made to enable rapid development for researchers.',
    long_description="GPJax aims to provide a low-level interface to Gaussian process models. Code is written entirely in Jax and Objax to enhance readability, and structured so as to allow researchers to easily extend the code to suit their own needs. When defining GP prior in GPJax, the user need only specify a mean and kernel function. A GP posterior can then be realised by computing the product of our prior with a likelihood function. The idea behind this is that the code should be as close as possible to the maths that we would write on paper when working with GP models.",
    install_requires=parse_requirements_file("requirements.txt"),
    # entry_points={
    #     "console_scripts": [
    #         "gpjax=gpjax.__main__:main",
    #     ]
    # },
)



from objax import TrainVar
from objax.typing import JaxArray
from objax.variable import reduce_mean
import jax.numpy as jnp
from typing import Optional, Callable
from .transforms import Transform, Softplus
from tensorflow_probability.substrates.jax import distributions as tfd


class Parameter(TrainVar):
    """
    Base parameter class. This is a simple extension of the `TrainVar` class in Objax that enables parameter transforms
    and, in the future, prior distributions to be placed on the parameter in question.
    """
    def __init__(self,
                 tensor: JaxArray,
                 reduce: Optional[Callable[[JaxArray],
                                           JaxArray]] = reduce_mean,
                 transform: Transform = Softplus(),
                 prior: tfd.Distribution = None):
        """
        Args:
            tensor: The initial value of the parameter
            reduce: A helper function for parallelisable calls.
            transform: The bijective transformation that should be applied to the parameter.
        """
        super().__init__(transform.forward(tensor), reduce)
        self.fn = transform
        self.prior = prior

    @property
    def untransform(self) -> JaxArray:
        """
        Return the paramerter's transformed valued that exists on constrained R.
        """
        return self.fn.backward(self.value)

    @property
    def log_density(self) -> JaxArray:
        """
        Return the log prior density of the parameter's constrained parameter value.
        """
        if self.prior is None:
            lpd = jnp.zeros_like(self.value)
        else:
            lpd = self.prior.log_prob(self.untransform)
        return lpd



from objax import Module
import jax.numpy as jnp


class MeanFunction(Module):
    def __init__(self, name: str = "Mean Function"):
        self.name = name

    def __call__(self, X: jnp.ndarray):
        raise NotImplementedError


class ZeroMean(MeanFunction):
    def __init__(self, name: str = "Zero Mean"):
        super().__init__(name=name)

    def __call__(self, X: jnp.ndarray) -> jnp.ndarray:
        return jnp.zeros((X.shape[0], 1), dtype=X.dtype)
